@@ -7,7 +7,33 @@ import haiku as hk
 
 import warnings
 
-from .nn import Representation, Prediction, Dynamic  # Example modules
+from .utils import scale_gradient, scalar_to_support, support_to_scalar, min_max
+
+
+def optimizer(init_value=0,
+              peak_value=2e-2, 
+              end_value=1e-3,
+              warmup_steps=1000,
+              transition_steps=10000,
+              decay_rate=0.8,
+              clip_by_global_norm=1.0
+              ):
+  scheduler = optax.warmup_exponential_decay_schedule(
+    init_value=init_value, 
+    peak_value=peak_value,
+    end_value=end_value,
+    warmup_steps=warmup_steps,
+    transition_steps=transition_steps,
+    decay_rate=decay_rate)
+  gradient_transform = optax.chain(
+      optax.clip_by_global_norm(clip_by_global_norm),  # Clip by the gradient by the global norm.
+      optax.scale_by_adam(),  # Use the updates from adam.
+      optax.scale_by_schedule(scheduler),  # Use the learning rate from the scheduler.
+      # Scale updates by -1 since optax.apply_updates is additive and we want to descend on the loss.
+      optax.scale(-1.0)
+  )
+  return gradient_transform
+
 
 class MuZero:
   r"""Muzero algorithm
@@ -20,45 +46,35 @@ class MuZero:
         The embedding dimension of hidden state `s`. Depends on the representation module. 
     num_actions: Any. 
         The maximum number of actions the policy can take. Depends on the prediction module and dynamic module.
-    representation_module: A class inherents hk.Module, 
+    representation_fn: A function initialized from a class which inherents hk.Module, 
         which takes raw observation `obs` from the environment as input and outputs the hidden state `s`.
         `s` will be the input of prediction_module and dynamic_module.
-    prediction_module: A class inherents hk.Module, 
+    prediction_fn: A function initialized from a class which inherents hk.Module, 
         which takes hidden state `s` as input and outputs prior logits `logits` and value `v` of the state.
-    dynamic_module: A class inherents hk.Module,
+    dynamic_fn: A function initialized from a class which inherents hk.Module,
         which takes hidden state `s` and action `a` as input and outputs reward `r` and next hidden state `ns`.
     policy: str, value in `['muzero', 'gumbel']`. Determines which muzero policy in mctx to use. 
     optimizer: Optimizer to update the parameters of `representation_module`, `prediction_module` and `dynamic_module`.
     discount: Any. Used for mctx.RecurrentFnOutput.
   """
   def __init__(self, 
-               embedding_dim,
-               num_actions,
-               representation_module,
-               prediction_module,
-               dynamic_module,
+               representation_fn,
+               prediction_fn,
+               dynamic_fn,
                policy='muzero',
-               optimizer = optax.adam(0.01),
-               discount: float = 0.99
+               optimizer = optimizer(),
+               discount: float = 0.99,
+               support_size: int = 10
                ):
-    self.repr_func = self._init_representation_func(representation_module, 
-                                                    embedding_dim) 
-    self.repr_func = hk.without_apply_rng(hk.transform(self.repr_func))
-
-    self.pred_func = self._init_prediction_func(prediction_module, 
-                                                num_actions)
-    self.pred_func = hk.without_apply_rng(hk.transform(self.pred_func))
-
-    self.dy_func = self._init_dynamic_func(dynamic_module, 
-                                           embedding_dim, num_actions)
-    self.dy_func = hk.without_apply_rng(hk.transform(self.dy_func))
+     
+    self.repr_func = hk.without_apply_rng(hk.transform(representation_fn))
+    self.pred_func = hk.without_apply_rng(hk.transform(prediction_fn))
+    self.dy_func = hk.without_apply_rng(hk.transform(dynamic_fn))
     
     self._policy = self._init_policy(policy)
     self._optimizer = optimizer 
     self._discount = discount
-
-    self.num_actions = num_actions
-    self.embedding_dim = embedding_dim
+    self._support_size = support_size
   
   def init(self, rng_key, sample_input):
     """Inits `representation`, `prediction` and `dynamic` functions and optimizer
@@ -99,6 +115,7 @@ class MuZero:
           with_value: bool = False,
           obs_from_batch: bool = False,
           num_simulations: int = 5,
+          temperature: float = 1.,
           *args, **kwargs):
     """Acts given environment's observations.
     
@@ -118,7 +135,7 @@ class MuZero:
     """
     if not obs_from_batch:
       obs = jnp.expand_dims(obs, axis=0)
-    plan_output, root_value = self._plan(self.params, rng_key, obs, num_simulations,
+    plan_output, root_value = self._plan(self.params, rng_key, obs, num_simulations, temperature,
                                         *args, **kwargs)
     root_value = root_value.item() if not obs_from_batch else root_value
     action = plan_output.action.item() if not obs_from_batch else plan_output.action
@@ -132,6 +149,7 @@ class MuZero:
     loss, grads = jax.value_and_grad(self._loss_fn)(self._params, batch, c)
     self._params, self._opt_state = self._update(self._params, self._opt_state, grads)
     loss_metric = {'loss': loss.item()}
+    
     return loss_metric
   
   def save(self, file):
@@ -154,10 +172,12 @@ class MuZero:
 
   def _plan(self, params, rng_key, obs,
            num_simulations: int = 5,
+           temperature: float = 1.,
            *args, **kwargs):
     root = self._root_inference(params, rng_key, obs)
     plan_output = self._policy(params, rng_key, root, self._recurrent_inference,
                                num_simulations=num_simulations,
+                               temperature=temperature,
                                *args, **kwargs)
     return plan_output, root.value
     
@@ -171,24 +191,31 @@ class MuZero:
   def _loss_fn(self, params, batch, c: float = 1e-4):
     loss = 0
     B, L, _ = batch.a.shape
+    batch.r = scalar_to_support(batch.r, self._support_size)
+    batch.Rn = scalar_to_support(batch.Rn, self._support_size)
     s = self._repr_apply(params['representation'], batch.obs[:, 0, :])
     # TODO: jax.lax.scan (or stay with fori_loop ?)
     def body_func(i, loss_s):
       loss, s = loss_s
       v, logits = self._pred_apply(params['prediction'], s)
+      # Appendix G, scale the gradient at the start of the dynamics function by 1/2 
+      s = scale_gradient(s, 0.5)
       r, s = self._dy_apply(params['dynamic'], s, batch.a[:, i, :].flatten())
       # losses: reward
-      loss_r = jnp.mean(optax.l2_loss(r, 
-                                      jax.lax.stop_gradient(batch.r[:, i, :])
-                                      ))
+      loss_r = jnp.mean(
+        optax.softmax_cross_entropy(r, 
+        jax.lax.stop_gradient(batch.r[:, i, :])
+        ))
       # losses: value
-      loss_v = jnp.mean(optax.l2_loss(v, 
-                                      jax.lax.stop_gradient(batch.Rn[:, i, :])
-                                      ))
+      loss_v = jnp.mean(
+        optax.softmax_cross_entropy(v, 
+        jax.lax.stop_gradient(batch.Rn[:, i, :])
+        ))
       # losses: action weights
-      loss_pi = jnp.mean(optax.softmax_cross_entropy(logits, 
-                                                     jax.lax.stop_gradient(batch.pi[:, i, :])
-                                                     ))
+      loss_pi = jnp.mean(
+        optax.softmax_cross_entropy(logits, 
+        jax.lax.stop_gradient(batch.pi[:, i, :])
+        ))
 
       loss += loss_r + loss_v + loss_pi 
       loss_s = (loss, s)
@@ -197,13 +224,10 @@ class MuZero:
     # Appendix G Training: "irrespective of how many steps we unroll for"
     loss /= L 
 
-    # L2 regularizer
-    # TODO: BUG all nan
-    # flatten_params, _ = jax.flatten_util.ravel_pytree(
-    #     jax.tree_map(lambda theta: jnp.linalg.norm(theta, ord=2), params)
-    #     )
-    
-    # loss += c * jnp.sum(flatten_params)
+    # L2 regulariser
+    l2_regulariser = 0.5 * sum(
+      jnp.sum(jnp.square(p)) for p in jax.tree_util.tree_leaves(params))
+    loss += c * jnp.sum(l2_regulariser)
     # print(f'loss2: {loss}')
     return loss
 
@@ -211,7 +235,7 @@ class MuZero:
   def _root_inference(self, params, rng_key, obs):
     s = self._repr_apply(params['representation'], obs)
     v, logits = self._pred_apply(params['prediction'], s)  
-    v = v.flatten()
+    v = support_to_scalar(jax.nn.softmax(v), self._support_size).flatten()
     root = mctx.RootFnOutput(
         prior_logits=logits,
         value=v,
@@ -223,7 +247,8 @@ class MuZero:
   def _recurrent_inference(self, params, rng_key, action, embedding):
     r, next_embedding = self._dy_apply(params['dynamic'], embedding, action)
     v, logits = self._pred_apply(params['prediction'], embedding)
-    r, v = r.flatten(), v.flatten()
+    r = support_to_scalar(jax.nn.softmax(r), self._support_size).flatten()
+    v = support_to_scalar(jax.nn.softmax(v), self._support_size).flatten()
     discount = jnp.ones_like(r) * self._discount
     recurrent_output = mctx.RecurrentFnOutput(
         reward=r,
@@ -256,23 +281,6 @@ class MuZero:
     else:
       warnings.warn(f"{policy} is not in ['muzero', 'gumbel'], uses muzero policy instead")
       policy_func = mctx.muzero_policy
-    return jax.jit(policy_func, static_argnums=(3, 4, ))
+    return jax.jit(policy_func, static_argnames=('temperature', 'recurrent_fn', 'num_simulations'), backend='cpu')
 
-  def _init_representation_func(self, representation_module, embedding_dim):
-    def representation_func(obs):
-      repr_model = representation_module(embedding_dim)
-      return repr_model(obs)
-    return representation_func
-  
-  def _init_prediction_func(self, prediction_module, num_actions):
-    def prediction_func(s):
-      pred_model = prediction_module(num_actions)
-      return pred_model(s)
-    return prediction_func
-
-  def _init_dynamic_func(self, dynamic_module, embedding_dim, num_actions):
-    def dynamic_func(s, a):
-      dy_model = dynamic_module(embedding_dim, num_actions)
-      return dy_model(s, a)
-    return dynamic_func 
 
